#include <stdint.h>
//#include <avr/pgmspace.h>

extern void printP(const char*);
extern void print(const char*);
extern void print(int i);

#define MACRO(A) do { A; } while(0)
#define MIN(a, b) ((a) < (b) ? (a) : (b))

//#define CHARS2INT(ptr) ntohl(*(uint32_t*)(ptr))
//#define INT2CHARS(ptr, val) MACRO( *(uint32_t*)(ptr) = htonl(val) )
#define CHARS2INT(ptr) (*(uint32_t*)(ptr))
#define INT2CHARS(ptr, val) MACRO( *(uint32_t*)(ptr) = (val) )

#define cbi(sfr,bit) (_SFR_BYTE(sfr) &= ~_BV(bit))


/******************************************************************************/

#ifndef DEGREE 
#define DEGREE  53                       /* the degree of the field polynomial */
#endif

#define MARGIN 3                                          /* don't touch this */
#define NUMWORDS ((DEGREE + MARGIN + 31) / 32)


/* the following type will represent bit vectors of length (DEGREE+MARGIN) */
typedef uint32_t bitstr_t[NUMWORDS];

typedef bitstr_t elem_t;           /* this type will represent field elements */
typedef bitstr_t exp_t;

void printBITSTR(const bitstr_t b, const char* label);

extern elem_t poly;                                      /* the reduction polynomial */
extern elem_t coeff_b, base_x, base_y;
extern exp_t base_order;

/******************************************************************************/


#define bitstr_clrbit(A, idx) MACRO( A[(idx) / 32] &= ~( ((uint32_t)1) << ((idx) % 32)) )
#define bitstr_copy(A, B) MACRO( memcpy(A, B, sizeof(bitstr_t)) )
#define bitstr_getbit(A, idx) ((A[(idx) / 32] >> ((idx) % 32)) & (uint32_t) 1)
#define bitstr_setbit(A, idx) MACRO( A[(idx) / 32] |= ((uint32_t)1) << ((idx) % 32) )
#define bitstr_clear(A) MACRO( memset(A, 0, sizeof(bitstr_t)) )
#define bitstr_swap(A, B) MACRO( bitstr_t h;  bitstr_copy(h, A); bitstr_copy(A, B); bitstr_copy(B, h) )
#define bitstr_is_equal(A, B) (! memcmp(A, B, sizeof(bitstr_t)) )

void bitstr_to_hex(char *s, const bitstr_t x);
int bitstr_sizeinbits(const bitstr_t x);
int bitstr_parse(bitstr_t x, const char *s);
void bitstr_import(bitstr_t x, const char *s);
int bitstr_is_clear(const bitstr_t x);
void bitstr_export(char *s, const bitstr_t x);
void bitstr_lshift(bitstr_t A, const bitstr_t B, int count);

/******************************************************************************/

#define field_add1(A) MACRO( A[0] ^= 1 )


void field_mult(elem_t z, const elem_t x, const elem_t y);
void field_add(elem_t z, const elem_t x, const elem_t y);
void field_invert(elem_t z, const elem_t x);

/******************************************************************************/


#define point_copy(x1, y1, x2, y2) MACRO( bitstr_copy(x1, x2); bitstr_copy(y1, y2) )
#define point_is_zero(x, y) (bitstr_is_clear(x) && bitstr_is_clear(y))

int is_point_on_curve(const elem_t x, const elem_t y);
void point_mult(elem_t x, elem_t y, const exp_t exp);
void point_double(elem_t x, elem_t y);


/******************************************************************************/

#define ECIES_OVERHEAD (8 * NUMWORDS + 8)

void ECIES_generate_key_pair(void);
void ECIES_encryption(char *msg, const char *text, int len, const char *Px, const char *Py);
int ECIES_decryption(char *text, const char *msg, int len, const char *privkey);
void ECIES_kdf(char *k1, char *k2, const elem_t Zx, const elem_t Rx, const elem_t Ry);
int ECIES_embedded_public_key_validation(const elem_t Px, const elem_t Py);
int ECIES_public_key_validation(const char *Px, const char *Py);

/******************************************************************************/

void XTEA_davies_meyer(char *out, const char *in, int ilen);
void XTEA_cbcmac(char *mac, const char *data, int size, const char *key);
void XTEA_ctr_crypt(char *data, int size, const char *key); 
void XTEA_encipher_block(char *data, const uint32_t *k);
void XTEA_init_key(uint32_t *k, const char *key);

/******************************************************************************/






