
#include "ecc.h"
#include "Arduino.h"

/******************************************************************************/
extern void generate_random(char * buffer, int length);

                               /* draw a random value 'exp' with 1 <= exp < n */
void get_random_exponent(exp_t exp)
{
  char buf[4 * NUMWORDS];
  int fh, r, s;
  do {
    generate_random(buf, 4*NUMWORDS);
    bitstr_import(exp, buf);
    for(r = bitstr_sizeinbits(base_order) - 1; r < NUMWORDS * 32; r++) 
    {
      bitstr_clrbit(exp, r);
    }
  } while(bitstr_is_clear(exp));
}


/******************************************************************************/

void ECIES_generate_key_pair(elem_t x, elem_t y, exp_t k)      /* generate a public/private key pair */
{
  get_random_exponent(k);
  point_copy(x, y, base_x, base_y);
  point_mult(x, y, k);
}

       /* check that a given elem_t-pair is a valid point on the curve != 'o' */
int ECIES_embedded_public_key_validation(const elem_t Px, const elem_t Py)
{
  return (bitstr_sizeinbits(Px) > DEGREE) || (bitstr_sizeinbits(Py) > DEGREE) ||
    point_is_zero(Px, Py) || ! is_point_on_curve(Px, Py) ? -1 : 1;
}

      /* same thing, but check also that (Px,Py) generates a group of order n */
int ECIES_public_key_validation(const char *Px, const char *Py)
{
  elem_t x, y;
  if ((bitstr_parse(x, Px) < 0) || (bitstr_parse(y, Py) < 0))
    return -1;
  if (ECIES_embedded_public_key_validation(x, y) < 0)
    return -1;
  point_mult(x, y, base_order);
  return point_is_zero(x, y) ? 1 : -1;
}

void ECIES_kdf(char *k1, char *k2, const elem_t Zx,     /* a non-standard KDF */
	       const elem_t Rx, const elem_t Ry)
{
  int bufsize = (3 * (4 * NUMWORDS) + 1 + 15) & ~15;
  char buf[bufsize];
  memset(buf, 0, bufsize);
  bitstr_export(buf, Zx);
  bitstr_export(buf + 4 * NUMWORDS, Rx);
  bitstr_export(buf + 8 * NUMWORDS, Ry);
  buf[12 * NUMWORDS] = 0; XTEA_davies_meyer(k1, buf, bufsize / 16);
  buf[12 * NUMWORDS] = 1; XTEA_davies_meyer(k1 + 8, buf, bufsize / 16);
  buf[12 * NUMWORDS] = 2; XTEA_davies_meyer(k2, buf, bufsize / 16);
  buf[12 * NUMWORDS] = 3; XTEA_davies_meyer(k2 + 8, buf, bufsize / 16);
}

                  /* ECIES encryption; the resulting cipher text message will be
                                            (len + ECIES_OVERHEAD) bytes long */
void ECIES_encryption(char *msg, const char *text, int len, 
		      const char *Px, const char *Py)
{
  printP(PSTR("\nencryption...\n"));
  
  elem_t Rx, Ry, Zx, Zy;
  char k1[16], k2[16];
  exp_t k;
  do {
    get_random_exponent(k);
    bitstr_parse(Zx, Px);
  printBITSTR(Zx,PSTR("pubx Zx"));
    bitstr_parse(Zy, Py);
    point_mult(Zx, Zy, k);
  printBITSTR(Zx,PSTR("k * Zx"));
    point_double(Zx, Zy);                           /* cofactor h = 2 on B163 */    
  printBITSTR(Zx,PSTR("2 * Zx"));
  } while(point_is_zero(Zx, Zy));
  
  point_copy(Rx, Ry, base_x, base_y);
  point_mult(Rx, Ry, k);

  printBITSTR(Zx,PSTR("Zx"));

  ECIES_kdf(k1, k2, Zx, Rx, Ry);  

  bitstr_export(msg, Rx);
  bitstr_export(msg + 4 * NUMWORDS, Ry);
  memcpy(msg + 8 * NUMWORDS, text, len);
  XTEA_ctr_crypt(msg + 8 * NUMWORDS, len, k1);
  XTEA_cbcmac(msg + 8 * NUMWORDS + len, msg + 8 * NUMWORDS, len, k2);
}

                                                          /* ECIES decryption */
int ECIES_decryption(char *text, const char *msg, int len, 
		     const char *privkey)
{
  printP(PSTR("\ndecryption...\n"));
  
  elem_t Rx, Ry, Zx, Zy;
  char k1[16], k2[16], mac[8];
  exp_t d;
  bitstr_import(Rx, msg);
  bitstr_import(Ry, msg + 4 * NUMWORDS);

  if (ECIES_embedded_public_key_validation(Rx, Ry) < 0) {
    printP(PSTR("ERROR: msg header not valid\n"));
    return -1;
  }
  
  bitstr_parse(d, privkey);
  
  point_copy(Zx, Zy, Rx, Ry);  
  point_mult(Zx, Zy, d);
  point_double(Zx, Zy);                             /* cofactor h = 2 on B163 */
  
  if (point_is_zero(Zx, Zy)) {
    printP(PSTR("ERROR: error priv key not valid \n"));
        return -1;
  }

  printBITSTR(Zx,PSTR("Zx"));
  ECIES_kdf(k1, k2, Zx, Rx, Ry);   
 
  XTEA_cbcmac(mac, msg + 8 * NUMWORDS, len, k2);
  if (memcmp(mac, msg + 8 * NUMWORDS + len, 8)) {
    printP( PSTR("ERROR: MAC not correct\n") );
    return -1;
  }
  memcpy(text, msg + 8 * NUMWORDS, len);
  XTEA_ctr_crypt(text, len, k1);
  return 1;
}




